/***************************************************************/
/*   Author: Erwan Loisant                                     */
/*   Date: September 2003                                      */
/*   Licence: General Public Licence                           */
/*                                                             */
/*   Contributors: Gerald Vogt, Todd Rudick, Mathieu Mangeot   */
/***************************************************************/

//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.

//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.

//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

///////////////////////////////////////////
//   GOBALS
///////////////////////////////////////////

if(!org) var org={};
if(!org.mozdev) org.mozdev={};
if(!org.mozdev.moji) org.mozdev.moji={};

org.mozdev.moji.main = function() {
    var pub={};

	const MOJI_RDF_ROOT = "urn:moji:words";
	const MOJI_NS = "http://moji.mozdev.org/rdf#";
	const MOJI_PREF_TOPIC = "mojiPrefUpdate";
	
	var RDFService = Components.classes["@mozilla.org/rdf/rdf-service;1"].getService().QueryInterface(Components.interfaces.nsIRDFService);
	var RDFContainerUtilsService = Components.classes["@mozilla.org/rdf/container-utils;1"].getService().QueryInterface(Components.interfaces.nsIRDFContainerUtils);
	
	var strbundle;
	
	var gMojiDS;
	var gMojiRoot;
	var gMojiRootContainer;
	
	var dicts;
	var kanjidicts;
	
	//var pub.kanjiIndex;
	var longestMatch;
	var defaultAction;
	
	var historyBox;
	
	var mojiInitialized = false;
	
	/***************************************************************/
	/*                Initialization functions                     */
	/***************************************************************/
	
	var prefObserver = {
	  observe : function(aSubject, aTopic, aData) {
	    // alert("observe: subject " + aSubject + ", topic " + aTopic + ", data = " + aData);
	    pub.initMoji();
	  }
	};
	
	var pref = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
	var mojiBranch = pref.getBranch("moji.");
	
	var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
	
	//////////////////////////////////////////////////////////////
	// initMoji
	//
	// PARAMETERS:
	//   nothing
	// RETURNS:
	//   nothing
	//
	// Load preferences and initialize global variables
	//////////////////////////////////////////////////////////////
	pub.initMoji = function()
	{
	  var dictorder = mojiBranch.getCharPref("dictorder");
	  var kanjidictorder = mojiBranch.getCharPref("kanjidictorder");
	  var deactdicts = mojiBranch.getCharPref("deactdicts");
	  var deactkanjidicts = mojiBranch.getCharPref("deactkanjidicts");
	
	  try {
	    longestMatch = mojiBranch.getBoolPref("longestMatch");
	  } catch ( e ) {
	    longestMatch = false;
	  }
	 
	  defaultAction = mojiBranch.getCharPref("defaultaction");
	  
	  pub.kanjiIndex = new Array();
	  pub.kanjiIndex["C"] = mojiBranch.getBoolPref("kanji.classicalchk");
	  pub.kanjiIndex["H"] = mojiBranch.getBoolPref("kanji.halpernchk");
	  pub.kanjiIndex["N"] = mojiBranch.getBoolPref("kanji.nelsonchk");
	  pub.kanjiIndex["V"] = mojiBranch.getBoolPref("kanji.newnelsonchk");
	  pub.kanjiIndex["DR"] = mojiBranch.getBoolPref("kanji.fatherjdrchk");
	  pub.kanjiIndex["DK"] = mojiBranch.getBoolPref("kanji.kanjilearnerchk");
	  pub.kanjiIndex["DO"] = mojiBranch.getBoolPref("kanji.pgochk");
	  pub.kanjiIndex["P"] = mojiBranch.getBoolPref("kanji.skipchk");
	  pub.kanjiIndex["U"] = mojiBranch.getBoolPref("kanji.unihexchk");
	  pub.kanjiIndex["I"] = mojiBranch.getBoolPref("kanji.tuttledicchk");
	  pub.kanjiIndex["IN"] = mojiBranch.getBoolPref("kanji.tuttlekanjichk");
	  pub.kanjiIndex["Q"] = mojiBranch.getBoolPref("kanji.fourchk");
	  pub.kanjiIndex["MN"] = mojiBranch.getBoolPref("kanji.moro1chk");
	  pub.kanjiIndex["MP"] = mojiBranch.getBoolPref("kanji.moro2chk");
	  pub.kanjiIndex["E"] = mojiBranch.getBoolPref("kanji.henshallchk");
	  pub.kanjiIndex["K"] = mojiBranch.getBoolPref("kanji.gakkenchk");
	  pub.kanjiIndex["L"] = mojiBranch.getBoolPref("kanji.heisigchk");
	  pub.kanjiIndex["O"] = mojiBranch.getBoolPref("kanji.pgo2chk");
	  pub.kanjiIndex["W"] = mojiBranch.getBoolPref("kanji.koreanchk");
	  pub.kanjiIndex["Y"] = mojiBranch.getBoolPref("kanji.pinyinchk");
	  pub.kanjiIndex["B"] = mojiBranch.getBoolPref("kanji.bushuchk");
	  pub.kanjiIndex["F"] = mojiBranch.getBoolPref("kanji.freqchk");
	  pub.kanjiIndex["G"] = mojiBranch.getBoolPref("kanji.jouyouchk");
	  pub.kanjiIndex["DB"] = mojiBranch.getBoolPref("kanji.busypeoplechk");
	  pub.kanjiIndex["DC"] = mojiBranch.getBoolPref("kanji.kanjiwaychk");
	  pub.kanjiIndex["DG"] = mojiBranch.getBoolPref("kanji.kodanshachk");
	  pub.kanjiIndex["DH"] = mojiBranch.getBoolPref("kanji.guidehensallchk");
	  pub.kanjiIndex["DJ"] = mojiBranch.getBoolPref("kanji.kanjicontextchk");
	  pub.kanjiIndex["DS"] = mojiBranch.getBoolPref("kanji.guidesakadechk");
	  pub.kanjiIndex["DT"] = mojiBranch.getBoolPref("kanji.tuttlecardschk");
	  pub.kanjiIndex["S"] = mojiBranch.getBoolPref("kanji.strokechk");
	
	  strbundle = window.document.getElementById("strings");
	  
	  // TODO: code cleanup of the initializations of edicts and kanjidicts. Most of the 
	  // preference handling code should be wrapped into one single function for either ones...
	  
	  var dictsBranch = pref.getBranch("moji.dicts.");
	  var hasDicts = false;
	  var missingDicts = "";
	  
	  dicts = new Array();
	  kanjidicts = new Array();
	
	  // First load all the word dictionaries
	    
	  if ( dictorder.length > 0 ) {
		var dictsarr = dictorder.split(",");
		var newdictorder = "";
		  
	    // Go through the dictorder preference and load each dictionary.
		for ( var i = 0; i < dictsarr.length; i++ )
		{
		    if ( dictsarr[i].length > 0 ) {
		        try {
		          // This creates the dictionary. If the dictionary does not exists
		          // we will catch the exception.
				  var edict = new org.mozdev.moji.dict.EDict(dictsarr[i]);
			    
			      // Push it on the stack
				  dicts.push(edict);
				  
				  // And rebuilt the new dictorder string.
				  if ( newdictorder.length > 0 ) 
				    newdictorder += ",";
	
				  newdictorder += dictsarr[i];
				    
				  hasDicts = true;
			    } catch ( e ) {
			      // We could not load the dictionary.
			      // This is either because the dictionary has been uninstalled or because
			      // of an other error like a file system error.
			      // If the dictionary has been uninstalled its preference branch would be missing
			      // and thus its preftype is PREF_INVALID. If it is not PREF_INVALID
			      // we alert the user.
			      if ( dictsBranch.getPrefType(dictsarr[i]) != dictsBranch.PREF_INVALID )
			        alert(strbundle.getString("dictopenerror") + " " + dictsarr[i] + ": " + e);
			      else {
			        if ( missingDicts.length > 0 )
			          missingDicts += ",";
			            
			        missingDicts += dictsarr[i];
			      }
			    }
			}
		}
		
		// If the dictorder changed it is because some dictionary has been uninstalled.
		if ( dictorder != newdictorder ) 
		{
		    dictorder = newdictorder;
	        mojiBranch.setCharPref("dictorder", newdictorder);
		}
	  }
	  
	  if ( kanjidictorder.length > 0 ) {
		var dictsarr = kanjidictorder.split(",");
	    var newkanjidictorder = "";
	    	  
		for ( var i = 0; i < dictsarr.length; i++ )
		{
			if ( dictsarr[i].length > 0 ) {
			  try {
				var kanjidic = new org.mozdev.moji.dict.KanjiDict(dictsarr[i]);
			    
				kanjidicts.push(kanjidic);
				
				if ( newkanjidictorder.length > 0 )
				  newkanjidictorder += ",";
	
	            newkanjidictorder += dictsarr[i];
				  
				hasDicts = true;
			  } catch ( e ) {
			    if ( dictsBranch.getPrefType(dictsarr[i]) != dictsBranch.PREF_INVALID )
			      alert(strbundle.getString("kanjidictopenerror") + " " + dictsarr[i] + ": " + e);
	            else {
		          if ( missingDicts.length > 0 )
		            missingDicts += ",";
			            
		          missingDicts += dictsarr[i];
		        }
			  }
			}
		}
		
		if ( kanjidictorder != newkanjidictorder )
		{
		    kanjidictorder = newkanjidictorder;
		    mojiBranch.setCharPref("kanjidictorder", newkanjidictorder);
		}
	  }
	  
	  var count = new Object();
	  var childArray = dictsBranch.getChildList("", count);
	  var hasNewDicts = false;
	  
	  for ( var i = 0; i < count.value; i++ ) {
	    var dictname = childArray[i];
	    
	    if ( dictname.indexOf(".") >= 0 )
	      continue;
	    
	    if ( ! dictsBranch.getBoolPref(dictname) )
	      continue;
	      
	    var re = new RegExp("(^|,)" + dictname + "(,|$)");
	    
	    if ( ! re.test( deactdicts ) && ! re.test( dictorder ) ) {
		  try {
			var edict = new org.mozdev.moji.dict.EDict(dictname);
	  	
			dicts.push(edict);
			hasDicts = true;
			hasNewDicts = true;
			
			if ( dictorder.length > 0 ) 
			  dictorder += ",";
	
	        dictorder += dictname;
		  } catch ( e ) {
			alert(strbundle.getString("dictopenerror") + " " + dictname + ": " + e); 
		  }
	    }
	  }
	  
	  if ( hasNewDicts ) {
	    mojiBranch.setCharPref("dictorder", dictorder);
	  }
	  
	  // Clean up uninstalled dictionaries from deactdicts.
	  deactArray = deactdicts.split(",");
	  var newdeact = "";
	  
	  for ( var i = 0; i < deactArray.length; i++ ) {
	    var dictname = deactArray[i];
	    
	    if ( dictname.length > 0 ) {
	      try {
	        if ( ! dictsBranch.getBoolPref(dictname) )
	          continue;
	        
	        if ( newdeact.length > 0 )
	          newdeact += ",";
	          
	        newdeact += dictname;
	      } catch ( e ) {
	        // not installed. forget about it.
	      }
	    }
	  }
	  
	  if ( deactdicts != newdeact ) {
	    mojiBranch.setCharPref("deactdicts", newdeact);
	  }
	  
	  dictsBranch = pref.getBranch("moji.kanjidicts.");
	  childArray = dictsBranch.getChildList("", count);
	  var hasNewKanjiDicts = false;
	  
	  for ( var i = 0; i < count.value; i++ ) {
	    var dictname = childArray[i];
	    
	    if ( dictname.indexOf(".") >= 0 )
	      continue;
	    
	    if ( ! dictsBranch.getBoolPref(dictname) )
	      continue;
	      
	    var re = new RegExp("(^|,)" + dictname + "(,|$)");
	    
	    if ( ! re.test( deactkanjidicts ) && ! re.test( kanjidictorder ) ) {
		  try {
			var kanjidic = new org.mozdev.moji.dict.KanjiDict(dictname);
			
			kanjidicts.push(kanjidic);
			hasDicts = true;
			hasNewKanjiDicts = true;
	
			if ( kanjidictorder.length > 0 ) 
			  kanjidictorder += ",";
	
			kanjidictorder += dictname;
		  } catch ( e ) {
			alert(strbundle.getString("kanjidictopenerror") + " " + dictname + ": " + e);
		  }
	    }
	  }
	  
	  if ( hasNewKanjiDicts ) {
	    mojiBranch.setCharPref("kanjidictorder", kanjidictorder);
	  }
	  
	  // Clean up uninstalled dictionaries from deactdicts.
	  deactArray = deactkanjidicts.split(",");
	  newdeact = "";
	  
	  for ( var i = 0; i < deactArray.length; i++ ) {
	    var dictname = deactArray[i];
	    
	    if ( dictname.length > 0 ) {
	      try {
	        if ( ! dictsBranch.getBoolPref(dictname) )
	          continue;
	        
	        if ( newdeact.length > 0 )
	          newdeact += ",";
	          
	        newdeact += dictname;
	      } catch ( e ) {
	        // not installed. forget about it.
	      }
	    }
	  }
	  
	  if ( deactkanjidicts != newdeact ) {
	    mojiBranch.setCharPref("deactkanjidicts", newdeact);
	  }
	  
	  if ( ! hasDicts ) {
	    alert(strbundle.getString("dictRequired"));
	    throw strbundle.getString("dictRequired");
	  }
	 
	  if ( hasNewDicts || hasNewKanjiDicts ) {
	    alert(strbundle.getString("newdicts"));
	  }
	  
	  if ( missingDicts.length > 0 ) {
	    alert(strbundle.getString("missingdicts") + " " + missingDicts);
	  }
	  
	  if ( ! mojiInitialized ) {
	    historyBox = window.document.getElementById('mojihistory');
	    observerService.addObserver(prefObserver, MOJI_PREF_TOPIC, false);
	  }
	
	  // Prepare RDF
	  var dirService = Components.classes['@mozilla.org/file/directory_service;1'].getService(Components.interfaces.nsIProperties);
	  var profileDir = dirService.get('ProfD', Components.interfaces.nsILocalFile);
	
	  //Create file object in the user's profile path.
	  var file = Components.classes['@mozilla.org/file/local;1'].createInstance(Components.interfaces.nsILocalFile);
	  file.initWithPath(profileDir.path);
	  file.append("moji.rdf");
	  //If it doesn't exist, write it.
	  
	  var newfile = !file.exists();
	  
	  if (newfile) {
	    // Initialize moji.rdf.
	    file.createUnique(0,0700);
	    var ostream = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
	    ostream.init(file, 12, 0700, 0);
	    var is = "<?xml version=\"1.0\"?>\n<RDF:RDF xmlns:NC=\"http://home.netscape.com/NC-rdf#\" xmlns:RDF=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">\n<RDF:Seq RDF:about=\"urn:moji:words\">\n</RDF:Seq>\n</RDF:RDF>\n";
	
	    ostream.write(is, is.length);
	    ostream.close();
	  } 
	
	  gMojiDS = RDFService.GetDataSourceBlocking('file://'+file.path);
	  gMojiRoot = RDFService.GetResource(MOJI_RDF_ROOT);
	  gMojiRootContainer = RDFContainerUtilsService.MakeSeq(gMojiDS, gMojiRoot);
	
	  historyBox.database.AddDataSource(gMojiDS);
	  historyBox.ref = MOJI_RDF_ROOT;
	  
	  historyBox.builder.rebuild();
	
	  mojiInitialized = true;
	}
	
	pub.closeMoji = function()
	{
	  observerService.removeObserver(prefObserver, MOJI_PREF_TOPIC);
	  
	  for ( var i = 0; i < dicts.length; i++ )
	    dicts[i].close();
	
	  for ( var i = 0; i < kanjidicts.length; i++)
	    kanjidicts[i].close();
	
	  gMojiDS.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
	  gMojiDS.Flush();
	}
	
	// Get selected text in HTML window
	function getSelection() {
	   var focusedWindow = document.commandDispatcher.focusedWindow;
	   var winWrapper = new XPCNativeWrapper(focusedWindow, 'document', 'getSelection()');
	   var selection = String(winWrapper.getSelection());
	
	   if ( selection == "" ) {
	     var focusedElement = document.commandDispatcher.focusedElement;
	     
	     if ( focusedElement instanceof HTMLInputElement 
	          || focusedElement instanceof HTMLTextAreaElement ) {
	       if ( focusedElement.selectionStart < focusedElement.selectionEnd ) {
	         selection = focusedElement.value.substring(focusedElement.selectionStart, focusedElement.selectionEnd);
	       } 
	     }
	   }
	
	   return selection;
	}
	
	function getWord()
	{
	  var word = getSelection();
	  
	  // remove leading and trailing white space.
	  word = word.replace(/^\s/g, "");
	  word = word.replace(/\s$/g, "");
	  
	  if (word.length == 0) {
	    var sbox = window.document.getElementById('mojitextbox').value;
	    sbox = sbox.replace(/\s/g, "");
	    
	    if ( sbox.length == 0 ) {
	      return;
	    }
	    
	    word = sbox;
	  }
	  
	  return word;
	}
	
	// lookup() is the default action
	pub.lookup = function (shiftKey)
	{
	  if ( defaultAction == "word" )
	    if ( shiftKey )
	      pub.lookupKanji();
	    else
	      pub.lookupWord();
	  else if ( defaultAction == "kanji" )
	    if ( shiftKey ) 
	      pub.lookupWord();
	    else
	      pub.lookupKanji();
	}
	
	pub.lookupKanji = function()
	{
	  if ( ! mojiInitialized ) 
	    pub.initMoji();
	
	  if ( kanjidicts.length < 1 ) {
	    alert(strbundle.getString("dictRequired"));
	    return;
	  }
	  
	  var txt = getWord();
	  
	  if ( txt == null || txt.length == 0 ) {
	    alert(strbundle.getString("atleast1kanji"));
	    return;
	  }
	
	  // var allentries = new Array();
	  var firstKanji;
	  
	  for ( var i = 0; i < txt.length; i++ ) {
	    var c = txt.charAt(i);
	    
	    var s = searched(c, "kanji");
	    
	    if ( s ) {
	      if ( firstKanji == null ) {
	        firstKanji = s;
	      }
	      continue;
	    }
	
	    var histentry = addSearch(c, "kanji");
	
	    if ( histentry != null ) {
	      // for ( var j = 0; j < histentry.dict.length; j++ )
	      //  for ( var k = 0; k < histentry.dict[j].entries.length; k++ ) 
	      //    allentries.push(histentry.dict[j].entries[k].txt);
	          
	      if ( firstKanji == null ) 
	        firstKanji = histentry;
	    }
	  }
	  
	  /* if ( allentries.length > 0 ) {
	    var s = "";
	  
	    for ( var i = 0; i < allentries.length; i++ ) 
	    {
	      if ( i > 0 )
	        s = s + "\r\n";
	        
	      s = s + allentries[i];
	    }
	    
	    alert("found " + allentries.length + " entries:\r\n" + s);
	  } else */
	  
	  if ( firstKanji == null ) {
	    alert(strbundle.getString("atleast1kanji"));
	    return;
	  } else 
	    historyBox.view.selection.select(historyBox.view.getIndexOfResource(RDFService.GetResource(MOJI_RDF_ROOT +":hist"+firstKanji)));
	}
	
	function clearDS()
	{
	  if ( gMojiDS == null )
	    return;
	    
	  var node = RDFService.GetResource(MOJI_RDF_ROOT);
	  _clearDSNode(node);
	
	  var arcs = gMojiDS.ArcLabelsIn(node);
	  
	  while ( arcs.hasMoreElements() ) {
	    var arc = arcs.getNext();
	    var source = gMojiDS.GetSource(arc, node, true);
	    gMojiDS.Unassert(source, arc, node);
	  }
	}
	
	function _clearDSNode(node)
	{
	  var arcs = gMojiDS.ArcLabelsOut(node);
	
	  while ( arcs.hasMoreElements() ) {
	    var arc = arcs.getNext();
	    var target = gMojiDS.GetTarget(node, arc, true);
	
	    try {
	      target.QueryInterface(Components.interfaces.nsIRDFResource);
	      _clearDSNode(target);
	    } catch ( e ) { }
	    
	    gMojiDS.Unassert(node, arc, target)
	  }
	}
	
	pub.clearHist = function ()
	{
	  historyBox.view.selection.clearSelection();
	  
	  clearDS();
	
	  gMojiRoot = RDFService.GetResource(MOJI_RDF_ROOT);
	  gMojiRootContainer = RDFContainerUtilsService.MakeSeq(gMojiDS, gMojiRoot);
	
	  window.document.getElementById('kanjiframe').setAttribute('src', "chrome://moji/locale/welcome.html");
	}
	
	pub.exportHist = function (how)
	{
	  var text = "";
	  
	  if ( how == "raw" ) {
	    for ( var i = 0; i < gMojiRootContainer.GetCount(); i++ ) {
	      var node = RDFService.GetResource(MOJI_RDF_ROOT + ":hist" + i);
	      text += getContainerFields(node, "raw", "%0A");
	    }
	
	    window.open("data:text/plain;charset=utf-8," + text, "History", "menubar,scrollbars,resizable,dialog=no");
	  } else if ( how == "html" ) {
	    for ( var i = 0; i < gMojiRootContainer.GetCount(); i++ ) {
	      if ( text != "" )
	        text = text + "%0A";
	      
	      text = text + getHtml(i, null, null);
	    }
	
	    window.open("data:text/html;charset=utf-8," + text, "History");
	  }  
	}
	
	pub.historySelect = function ()
	{
	  var s = historyBox.view.selection.currentIndex;
	
	  if ( s < 0 )
	    return;
	  
	  historyBox.treeBoxObject.ensureRowIsVisible(s);
	  
	  var id = historyBox.view.getCellText(s, historyBox.columns.getNamedColumn("index"));
	
	  var histid = id.match(/^hist([0-9]+)(\.([0-9]+))?(\.([0-9]+))?/);
	
	  if ( histid[1] == null ) {
	    throw "regexp does not return histentno";
	    return;
	  }
	
	  var histno = histid[1];
	  var dictno = histid[3];
	  var matchno = histid[5];
	
	  if ( histno == null || (histno < 0) || (histno >= gMojiRootContainer.GetCount()) ) {
	    throw "Histno " + histno + " out of bounds (0:" + gMojiRootContainer.GetCount() + ")";
	  }
	  
	  var html = getHtml(histno, dictno, matchno); 
	  window.document.getElementById('kanjiframe').setAttribute('src', "data:text/html;charset=utf-8," + html);
	}
	
	// lookupWord()
	//
	// search for a word in the word dictionaries.
	//
	pub.lookupWord = function ()
	{
	  if ( ! mojiInitialized ) 
	    pub.initMoji();
	  
	  if ( dicts.length < 1 ) {
	    alert(strbundle.getString("dictRequired"));
	    return;
	  }
	  
	  // O.K. figure out which word we have to look for.
	  var txt = getWord();
	
	  // nothing to do...
	  if ( txt == null || txt.length == 0 ) {
	    alert(strbundle.getString("selectword"));
	    return;
	  }
	  
	  var histselitem;
	  var rest = "";
	  
	  while ( txt.length > 0 ) {
	    // check if we have searched for this term before. If yes, just select the corresponding top entry
	    // in the history box which will show the entry in turn.
	    var s = searched(txt, "word");
	    
	    if ( s ) {
	      if ( histselitem == null ) 
	        histselitem = s;
	      
	      txt = rest;
	      rest = "";
	      continue;
	    }
	
	    var histentry = addSearch(txt, "word");
	    
	    if ( histentry != null && histselitem == null ) {
	      histselitem = histentry;
	    }
	    
	    if ( longestMatch ) {
	      if ( histentry == null ) {
	        if ( txt.length > 1 ) {
		  if ( txt.length > 100 ) {
	            // if the text is longer than 100 characters we cut it
	            // to length 100 here to limit endless searches.
	            alert(strbundle.getString("toolong"));
	            txt = txt.substr(0, 100);
		    rest = "";
	          } else {
	            rest = txt.substr(txt.length - 1) + rest;
	            txt = txt.substr(0, txt.length - 1);
	          }
	        } else {
	          txt = rest;
	          rest = "";
	        }
	      } else {
	        txt = rest;
	        rest = "";
	      }
	    } else {
	      txt = "";
	    }
	    
	    // DEBUG debugging code to dump all the matches in the history entry.
	    /* if ( histentry != null ) {
	      var s = "";
	      var c = 0;
	    
	      for ( var i = 0; i < histentry.dict.length; i++ ) 
	      {
	        for ( var j = 0; j < histentry.dict[i].entries.length; j++ )
	        {
	          if ( c > 0 )
	            s = s + "\r\n";
	            
	          c = c + 1;
	          
	          s = s + histentry.dict[i].entries[j].txt;
	          
	          if ( histentry.dict[i].entries[j].reason ) 
	            s = s + " [" + histentry.dict[i].entries[j].reason + "]";
	        }
	      }
	      
	      alert("found " + c + " entries in " + histentry.dict.length + " dictionaries:\r\n" + s);
	    } else
	      alert("no entries found");
	    */ 
	  }
	  
	  if ( histselitem == null ) {
	        var sorry = "<html>" +
	                '<head><link rel="stylesheet" type="text/css" href="chrome://moji/skin/output.css" /></head>' +
	                "<body><div style='text-align: center'><p>"+ strbundle.getString("sorry") +"<br/>" +
		        "<span class='lookedword'>" + txt + "</span><br/>" +
			strbundle.getString("notfound")+"</p><br/>" +
			"<span class='kaomoji'>m(_ _)m</span>" +
			"</body></html>";
	    window.document.getElementById('kanjiframe').setAttribute('src', "data:text/html;charset=utf-8," + sorry);
	  } else {
	    // and select it which will have the entry rendered in the sidebox.
	    historyBox.view.selection.select(historyBox.view.getIndexOfResource(RDFService.GetResource(MOJI_RDF_ROOT +":hist"+histselitem)));
	  }
	}
	
	// structure of histentry:
	// histentry = Object();
	
	// histentry.searchword = string, searchword representing this entry
	// histentry.cls = string, class of entry: either "kanji" or "word"
	// histentry.dict = Array(), one dict entry for each dictionary active, in the configured order
	
	//  histentry.dict[].name = string, name of dictionary (e.g. "mojien")
	//  histentry.dict[].firstMeaning = string, extracted first meaning of the first entry.
	
	//  histentry.dict[].entries = Array() of Object(), representing each match in the dictionary
	
	//   histentry.dict[].entries[].txt = string, raw dictionary entry from the kanjidic or edict file.
	//   histentry.dict[].entries[].html = string, HTML representation of this entry
	//   histentry.dict[].entries[].firstMeaning = string, first meaning of this entry.
	
	//             *** Word dictionary ***
	//   histentry.dict[].entries[].reason = reason for conjugated match
	
	//             *** Kanji dictionary ***
	//   histentry.dict[].entries[].meanings[] = string, raw meaning entries (between slashes)
	//   histentry.dict[].entries[].key = string, kanji (first field of line)
	//   histentry.dict[].entries[].jiscode = integer, JIS code of kanji
	//   histentry.dict[].entries[].reading = Array() of strings, containing the readings of the kanji
	//   histentry.dict[].entries[].nameReading = Array() of strings, containing the readings of the name
	//   histentry.dict[].entries[].radicalReading = Array() of strings, containing the readings of the radical
	//   histentry.dict[].entries[].info = Object(), associative object, containing all other fields of the kanjidic entry by key
	
	
	// addSearch does a search for txt in the dict class cls.
	// cls can be "word" or "kanji" at this time.
	// addSearch returns the index of the history entry or null if nothing has been found.
	function addSearch(txt, cls)
	{
	  // O.K. txt contains the text we are searching for. Prepare the new history entry. 
	  var histentry = new Object();
	  var searchdicts;
	  
	  // remember the searchword and which class we are in.
	  histentry.searchword = txt;
	  histentry.cls = cls;
	
	  if ( cls == "word" ) 
	    searchdicts = dicts;
	  else if ( cls == "kanji" ) 
	    searchdicts = kanjidicts;
	  else
	    throw "Unknown class \"" + cls + "\"";
	      
	  // and prepare the array for the dictionaries.
	  histentry.dict = new Array();
	  
	  // now we run through all active dictionaries and search for the word in each of them.
	  for ( var i = 0; i < searchdicts.length; i++ ) {
	    // get all matches and derived/conjugated matches (if conjugation file is present)
	    // from the dictionary.
	    
	    var e = searchdicts[i].search(txt);
	    if ( e != null && e.length > 0 ) {
	      // O.K. we have found something in this dictionary. prepare the dictionary entry.
	      var histentdict = new Object();
	      // save the name of the dictionary of later.
	      histentdict.name = searchdicts[i].name;
	      
	      // now we have to process each of the found entries, generate the html code
	      for ( var j = 0; j < e.length; j++ ) {
	        var fm = searchdicts[i].getFirstMeaning(e[j].txt);
	
	        // this is the first match in this dictionary thus the first meaning of this match
	        // is also considered as first meaning for the whole dictionary.
	        if ( j == 0 ) {
	          histentdict.firstMeaning = fm;
	        }
	        
	        // create the html and remember the first meaning for this match.
	        e[j].html = searchdicts[i].formatEntry(e[j]);
	        e[j].firstMeaning = fm;
	      }
	      
	      // O.K. almost done. We need a first meaning for the whole search as well, which 
	      // we establish here if this is actually the first dictionary we found matches in.
	      if ( histentry.dict.length == 0 ) {
	        histentry.firstMeaning = histentdict.firstMeaning;
	      }
	
	      // save the new object and push it onto the dict array in the history entry.
	      histentdict.entries = e;
	      histentry.dict.push(histentdict);
	    }
	  }
	  
	  // if we found no matches the dicts array in the history entry is empty.
	  if ( histentry.dict.length == 0 )
	  {
	    return null;
	  } else {
	    // Found matches.
	    // the next index should depend on the highest existing one and not depend
	    // on a completely sequential order with no gaps. (gaps don't exist at the 
	    // moment but may in the future when it is possible to delete single entries in the
	    // history)
	    var ind = gMojiRootContainer.GetCount();
	    
	    // add the entry to the tree.
	    addHistoryEntry(histentry, ind);
	    
	    return ind;
	  }
	}
	
	function searched(txt, cls)
	{
	  var classProp = RDFService.GetResource(MOJI_NS + 'class')
	  var sources = gMojiDS.GetSources(RDFService.GetResource(MOJI_NS + 'word'), RDFService.GetLiteral(txt), true);
	  
	  while ( sources.hasMoreElements() ) {
	    var element = sources.getNext();
	    
	    var target = gMojiDS.GetTarget(element, classProp, true);
	    target.QueryInterface(Components.interfaces.nsIRDFLiteral);
	    
	    if ( target.Value == cls ) {
	      target = gMojiDS.GetTarget(element, RDFService.GetResource(MOJI_NS + 'id'), true);
	      target.QueryInterface(Components.interfaces.nsIRDFLiteral);
	      
	      var histid = target.Value.match(/^hist([0-9]+)(\.([0-9]+))?(\.([0-9]+))?/);
	      
	      if ( histid[1] == null )
	        throw "unknown histid '" + target.Value + "' in datasource";
	        
	      return histid[1];
	    }
	  }
	  
	  return null;
	}
	
	function getContainerFields(resource, prop, term)
	{
	  var txt = "";
	  
	  if ( RDFContainerUtilsService.IsSeq(gMojiDS, resource) ) {
	    var container = RDFContainerUtilsService.MakeSeq(gMojiDS, resource);
	    
	    if ( container.GetCount() == 0 ) 
	      return txt;
	      
	    var elements = container.GetElements();
	  
	    while ( elements.hasMoreElements() ) {
	      var element = elements.getNext();
	      
	      txt += getContainerFields(element, prop, term);
	    }
	  } else {
	    var target = gMojiDS.GetTarget(resource, RDFService.GetResource(MOJI_NS + prop), true);
	    target.QueryInterface(Components.interfaces.nsIRDFLiteral);
	    txt += target.Value;
	    if ( term )
	      txt += term;
	  }
	  
	  return txt;
	}
	
	// getHtml returns the html code for a selected history entry.
	// histent is the actual history entry which has been selected.
	// dictno is the dictionary index in case a sub entry has been selected in the tree.
	// matchno is the match in the dictionary dictno that has been selected.
	function getHtml(histent, dictno, matchno)
	{
	  var html = '<html>\n<head><link rel="stylesheet" type="text/css" href="chrome://moji/skin/output.css" /></head>\n<body>\n';
	  // the iframe only compares some xxx first bytes of the src tag to decide whether or not
	  // the contents has changed and needs to be repainted. Therefore switching between a tree node
	  // and its direct parent may result in no change as the first entry is always the identical
	  // parent node. For this reason we just a random number here in the beginning to make sure
	  // that (most likely ;-) each entry is unique.
	  html += "<!-- " + Math.random() + " -->";
	  
	  var ind = histent;
	  
	  if ( dictno ) {
	    ind = ind + "." + dictno;
	    
	    if ( matchno ) {
	      ind = ind + "." + matchno;
	    }
	  }
	    
	  var node = RDFService.GetResource(MOJI_RDF_ROOT + ":hist" + ind);
	  html += getContainerFields(node, "html");
	      
	  html += "</body></html>";
	  
	  return html;
	}
	
	// add the necessary items to the history tree.
	// - histent is the moji history entry to be added.
	// - ind defined the new index of the entry to be used.
	//
	// This function is a little bit tricky as the structure of the tree changes
	// depending on the number of matches found. The basic structure is:
	// + word dict1.match1
	// +-+ word dict1.match1
	// +-+-+ word dict1.match1
	// +-+-+ word dict1.match2
	// +-+ word dict2.match1
	// +-+-+ word dict2.match1
	// +-+-+ word dict2.match2
	// i.e. a higher level entry always duplicates the entry information of the first child therefore
	//   if you have the subtree hidden you always see the first match meaning on the subtree trunk.
	//
	// However, for optimization and reduction of unnecessary subtrees, no subtree will be added if
	// it would contain only one entry. That means if there is only one match in a dictionary the 
	// dictionary entry have no subtree and it represents itself the only match found. Likewise,
	// If matches are only found in one dictionary there won't be the intermediate dict1, dict2, ...
	// structure but instead the dictionary matches will be added as subtree to the trunk.
	// Finally, if there is actually only one match in one dictionary the trunk itself will 
	// represent this match and there won't be a subtree to this search word.
	//
	function addHistoryEntry(histent, ind)
	{
	  var currentTop;
	  var source = RDFService.GetResource(MOJI_RDF_ROOT + ":" + "hist" + ind);
	  
	  gMojiDS.Assert(source, RDFService.GetResource(MOJI_NS + 'id'), RDFService.GetLiteral("hist" + ind), true);
	  gMojiDS.Assert(source, RDFService.GetResource(MOJI_NS + 'class'), RDFService.GetLiteral(histent.cls), true);
	  gMojiDS.Assert(source, RDFService.GetResource(MOJI_NS + 'word'), RDFService.GetLiteral(histent.searchword), true);
	  gMojiDS.Assert(source, RDFService.GetResource(MOJI_NS + 'meaning'), RDFService.GetLiteral(histent.dict[0].entries[0].firstMeaning), true);
	  gMojiDS.Assert(source, RDFService.GetResource(MOJI_NS + 'reason'), RDFService.GetLiteral(histent.dict[0].entries[0].reason || ""), true);
	  gMojiDS.Assert(source, RDFService.GetResource(MOJI_NS + 'dictionary'), RDFService.GetLiteral(histent.dict[0].name || ""), true);
	
	  if ( ( histent.dict.length == 1 ) && ( histent.dict[0].entries.length == 1 ) ) {
	    gMojiDS.Assert(source, RDFService.GetResource(MOJI_NS + 'html'), RDFService.GetLiteral(histent.dict[0].entries[0].html), true);
	    gMojiDS.Assert(source, RDFService.GetResource(MOJI_NS + 'raw'), RDFService.GetLiteral(histent.dict[0].entries[0].txt), true);
	  }
	  
	  gMojiRootContainer.AppendElement(source);
	
	  var sourceContainer;
	  var currentTopContainer;
	  var currentTop = source;
	  
	  for ( var i = 0; i < histent.dict.length; i++ ) {
	    // if there is more than one dictionary with matches we have to create
	    // a subtree for the dictionaries.
	    var dictIndex = "hist"+ind+"."+i.toString();
	    if ( histent.dict.length > 1 ) {
	       var dictNode = RDFService.GetResource(MOJI_RDF_ROOT + ":" + dictIndex);
	
	       gMojiDS.Assert(dictNode, RDFService.GetResource(MOJI_NS + 'id'), RDFService.GetLiteral(dictIndex), true);
	       gMojiDS.Assert(dictNode, RDFService.GetResource(MOJI_NS + 'class'), RDFService.GetLiteral(histent.cls), true);
	       gMojiDS.Assert(dictNode, RDFService.GetResource(MOJI_NS + 'word'), RDFService.GetLiteral(histent.searchword), true);
	       gMojiDS.Assert(dictNode, RDFService.GetResource(MOJI_NS + 'meaning'), RDFService.GetLiteral(histent.dict[i].entries[0].firstMeaning), true);
	       gMojiDS.Assert(dictNode, RDFService.GetResource(MOJI_NS + 'reason'), RDFService.GetLiteral(histent.dict[i].entries[0].reason || ""), true);
	       gMojiDS.Assert(dictNode, RDFService.GetResource(MOJI_NS + 'dictionary'), RDFService.GetLiteral(histent.dict[i].name || ""), true);
	
	       if ( histent.dict[i].entries.length == 1 ) {
	         gMojiDS.Assert(dictNode, RDFService.GetResource(MOJI_NS + 'html'), RDFService.GetLiteral(histent.dict[i].entries[0].html), true);
	         gMojiDS.Assert(dictNode, RDFService.GetResource(MOJI_NS + 'raw'), RDFService.GetLiteral(histent.dict[i].entries[0].txt), true);
	       }
	       
	       if ( ! sourceContainer )
	        sourceContainer = RDFContainerUtilsService.MakeSeq(gMojiDS, source);
	
	       sourceContainer.AppendElement(dictNode);
	
	       currentTop = dictNode;
	       currentTopContainer = null;
	    }
	    
	    if (histent.dict[i].entries.length > 1) {
	      for ( var j = 0; j < histent.dict[i].entries.length; j++ ) {
	        var entryIndex = dictIndex + "." + j.toString();
	        var entryNode = RDFService.GetResource(MOJI_RDF_ROOT + ":" + entryIndex);
	
	        gMojiDS.Assert(entryNode, RDFService.GetResource(MOJI_NS + 'id'), RDFService.GetLiteral(entryIndex), true);
	        gMojiDS.Assert(entryNode, RDFService.GetResource(MOJI_NS + 'class'), RDFService.GetLiteral(histent.cls), true);
	        gMojiDS.Assert(entryNode, RDFService.GetResource(MOJI_NS + 'word'), RDFService.GetLiteral(histent.searchword), true);
	        gMojiDS.Assert(entryNode, RDFService.GetResource(MOJI_NS + 'meaning'), RDFService.GetLiteral(histent.dict[i].entries[j].firstMeaning), true);
	        gMojiDS.Assert(entryNode, RDFService.GetResource(MOJI_NS + 'reason'), RDFService.GetLiteral(histent.dict[i].entries[j].reason || ""), true);
	        gMojiDS.Assert(entryNode, RDFService.GetResource(MOJI_NS + 'dictionary'), RDFService.GetLiteral(histent.dict[i].name || ""), true);
	        gMojiDS.Assert(entryNode, RDFService.GetResource(MOJI_NS + 'html'), RDFService.GetLiteral(histent.dict[i].entries[j].html), true);
	        gMojiDS.Assert(entryNode, RDFService.GetResource(MOJI_NS + 'raw'), RDFService.GetLiteral(histent.dict[i].entries[j].txt), true);
	
	        if ( ! currentTopContainer )
	          currentTopContainer = RDFContainerUtilsService.MakeSeq(gMojiDS, currentTop);
	        
	        currentTopContainer.AppendElement(entryNode);
	       }
	     }
	   }
	}
	
	pub.showMojiTooltip = function (tipElement)
	{
	  if ( ! tipElement )
	    return false;
	    
	  var titleText = tipElement.getAttribute("title");
	  
	  if ( !titleText )
	    return false;
	    
	  var tipNode = document.getElementById("mojiTooltip");
	  tipNode.setAttribute("label", titleText);
	  return true;
	}
	
	return pub;
}();